# This file is part of gcaff
# Copyright (C) 2013 Fraser Tweedale
#
# gcaff is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
import shutil
import subprocess
import tempfile

logger = logging.getLogger(__name__)

(
    VALIDITY_NEW,
    VALIDITY_INVALID,
    VALIDITY_DISABLED,
    VALIDITY_REVOKED,
    VALIDITY_EXPIRED,
    VALIDITY_UNKNOWN,
    VALIDITY_UNDEFINED,
    VALIDITY_VALID,
    VALIDITY_MARGINAL,
    VALIDITY_FULL,
    VALIDITY_ULTIMATE,
) = range(11)

SIGNABLE = {
    VALIDITY_NEW: False,
    VALIDITY_INVALID: False,
    VALIDITY_DISABLED: False,
    VALIDITY_REVOKED: False,
    VALIDITY_EXPIRED: False,
    VALIDITY_UNKNOWN: True,
    VALIDITY_UNDEFINED: True,
    VALIDITY_VALID: True,
    VALIDITY_MARGINAL: True,
    VALIDITY_FULL: True,
    VALIDITY_ULTIMATE: True,
}

VALIDITY = {
    b'o': VALIDITY_NEW,
    b'i': VALIDITY_INVALID,
    b'd': VALIDITY_DISABLED,
    b'r': VALIDITY_REVOKED,
    b'e': VALIDITY_EXPIRED,
    b'-': VALIDITY_UNKNOWN,
    b'q': VALIDITY_UNDEFINED,
    b'n': VALIDITY_VALID,
    b'm': VALIDITY_MARGINAL,
    b'f': VALIDITY_FULL,
    b'u': VALIDITY_ULTIMATE,
}

ALGORITHM = {
    b'1':  'RSA',
    b'16': 'Elgamal (encrypt only)',
    b'17': 'DSA',
    b'20': 'Elgamal (sign and encrypt)',
}


class Uid(object):
    def __init__(self, validity, data):
        self._validity = validity
        self.data = data

    def validity(self):
        return VALIDITY.get(self._validity, VALIDITY_UNKNOWN)

    def signable(self):
        return SIGNABLE[self.validity()]


class TextUid(Uid):
    def name(self):
        return self._extract_parts()[0]

    def email(self):
        return self._extract_parts()[1]

    def _extract_parts(self):
        """Extract name and email address from the given uid."""
        match = re.match(r'(.*) <(.*?)>$', self.data)
        return match.group(1, 2) if match else (self.data, None)

    def __str__(self):
        return self.data


class ImageUid(Uid):
    def name(self):
        return '[image]'

    def email(self):
        return None

    def __str__(self):
        return '[jpeg image of size {}]'.format(len(self.data))


class Key(object):
    def __init__(self, length, algorithm, fingerprint, uids):
        self.length = length
        self.algorithm = algorithm
        self.fingerprint = fingerprint
        self.uids = uids

    def __eq__(self, other):
        return self.fingerprint == other.fingerprint

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def keyid(self):
        """Return the long key ID."""
        return self.fingerprint[-16:]

    def names(self):
        """Return a list of all names on non-image UIDs."""
        names = (
            uid.name() for uid in self.uids
            if isinstance(uid, TextUid) and uid.signable()
        )
        return filter(None, names)

    def name(self):
        """Get the first name on this key.

        Return a list containing the first name found on a
        valid UID, or the empty list if no such name is found.

        """
        return self.names()[:1]

    def emails(self):
        """Return a list of all valid email addresses on the key."""
        emails = (uid.email() for uid in self.uids if uid.signable())
        return filter(None, emails)

    def email(self):
        """Get a usable email address for this key.
        
        Return a list containing the first email address found on a
        valid UID, or the empty list if no such address is found.

        """
        return self.emails()[:1]

    def uid_email(self, i):
        """Get an email address for the given UID.

        If the given UID does not have an email address, return a
        list of all email addresses on the key, in the hope that
        one of them will suffice for contacting the emailless UID.

        """
        email = self.uids[i].email()
        return [email] if email else self.emails()

    def human_algorithm(self):
        return ALGORITHM.get(self.algorithm, '<unknown algorithm>')

    def human_fingerprint(self):
        fpr = self.fingerprint
        return ' '.join(fpr[i:i + 4] for i in xrange(0, len(fpr), 4))


class GnuPG(object):
    """GnuPG wrapper.

    Leaving ``homedir`` unspecified uses the "normal" keyring, i.e.
    GNUPGHOME if set, else the GnuPG default of ``$HOME/.gnupg``.

    The secret keyring from the "normal" homedir is always used, so
    signing keys do not have to be moved around.

    """
    def __init__(self, homedir=None):
        # determine the user's normal GNUPGHOME, where we expect
        # to find signing keys
        self.homedir = homedir \
            or os.environ.get('GNUPGHOME') \
            or os.path.join(os.environ['HOME'], '.gnupg')
        secdir = os.environ.get('GNUPGHOME') \
            or os.path.join(os.environ['HOME'], '.gnupg')
        self.secret_keyring = os.path.join(secdir, 'secring.gpg')

    def _popen(self, args):
        #homedir = ['--homedir', self.homedir] if self.homedir else []
        args = [
            'gpg',
            '--no-tty',
            '--no-auto-check-trustdb',
            '--secret-keyring', self.secret_keyring,
        #] + homedir + args
        ] + args
        logger.info('execute gpg: GNUPGHOME={} {}'.format(self.homedir, args))
        return subprocess.Popen(
            args,
            env=dict(os.environ, GNUPGHOME=self.homedir),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

    def import_keys(self, data, minimal=False):
        """Import keys from the given data.

        ``minimal``
            Do a minimal import.  Defaults to ``False``.

        """
        logger.info('import')
        args = ['--import']
        if minimal:
            args += ['--import-options', 'import-minimal']
        p = self._popen(args)
        stdout, stderr = p.communicate(data)
        logger.debug('gpg --import exited code {!r}'.format(p.returncode))

    def export_keys(self, keyids=None, minimal=False):
        """Export the given keys, returning the armoured data."""
        logger.info('export; keyids: {!r}'.format(keyids))
        opts = 'export-minimal' if minimal \
            else 'no-export-minimal,no-export-clean'
        p = self._popen([
            '--armor',
            '--export',
            '--export-options', opts,
        ] + (keyids or []))
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            raise RuntimeError('gpg --export returned nonzero')
        return stdout

    def encrypt(self, keyid, data):
        """Encrypt ``data`` to the given ``keyid`` (with armour)."""
        logger.info('encrypt; recipient: {}'.format(keyid))
        p = self._popen([
            '--armor',
            '--trust-model', 'always',
            '--encrypt',
            '--no-encrypt-to',
            '--recipient', keyid
        ])
        stdout, stderr = p.communicate(data)
        if p.returncode != 0:
            raise RuntimeError(
                'gpg --encrypt returned {}'.format(p.returncode)
            )
        return stdout

    def list_secret_keys(self, keyids=None):
        keys = self._list_keys(keyids, secret=True)
        return self._list_keys([k.keyid for k in keys])

    def list_keys(self, keyids=None):
        return self._list_keys(keyids)

    def _list_keys(self, keyids=None, secret=False):
        """Retrieve the given key."""
        # FIXME secret keys are not displaying validity
        wstat_fd, rstat_fd = os.pipe()
        wattr_fd, rattr_fd = os.pipe()
        p = self._popen([
            '--status-fd', str(wstat_fd),
            '--attribute-fd', str(wattr_fd),
            '--fingerprint',
            '--with-colons',
            '--list-secret-keys' if secret else '--list-keys',
            '--list-options', 'show-uid-validity,show-unusable-uids',
        ] + (keyids or []))
        stdout, stderr = p.communicate()
        os.close(wstat_fd)
        os.close(wattr_fd)
        with os.fdopen(rstat_fd) as f:
            statout = f.read()
        images = self._get_images_from_key(statout, rattr_fd)
        os.close(rattr_fd)

        keys = []
        seen = 0
        kwargs = {'uids': []}
        for line in stdout.splitlines():
            if line.startswith('sec' if secret else 'pub'):
                if seen:
                    keys.append(Key(**kwargs))
                    kwargs = {'uids': []}
                _, _, kwargs['length'], kwargs['algorithm'] \
                    = line.split(':')[:4]
                seen += 1
            if line.startswith('fpr'):
                kwargs['fingerprint'] = line.split(':')[9]
            elif line.startswith('uid'):
                _, validity, _, _, _, _, _, _, _, data = line.split(':')[:10]
                kwargs['uids'].append(TextUid(validity, data))
            elif line.startswith('uat'):
                # if uat lines that are NOT jpeg images are listed,
                # this code is wrong and needs fixing.  But I have
                # never seen such a case.
                #
                _, validity = line.split(':')[:2]
                kwargs['uids'].append(ImageUid(validity, data=images.pop(0)))
        if seen:
            keys.append(Key(**kwargs))
        return keys

    @staticmethod
    def _get_images_from_key(statout, fd):
        images = []
        for line in statout.splitlines():
            match = re.match(r'''
                    \[GNUPG:\]\ ATTRIBUTE
                    \ \w+                   # fingerprint
                    \ (?P<octets>\d+)
                    \ 1                     # type (1 == image)
                    \ (?P<index>\d+)
                    \ (?P<count>\d+)
                    \                       # ... don't care about the rest
                ''', line, re.VERBOSE)
            if match:
                data = os.read(fd, int(match.group('octets')))
                if not data[:4] == '\x10\x00\x01\x01':
                    raise RuntimeError('invalid data in attribute packet')
                if match.group('index') == '1':
                    images.append(data[16:])
                else:
                    images[-1] += data[16:]
        return images

    STATE_MINIMIZE, STATE_INIT, STATE_UID, STATE_SIGN, STATE_SAVE = range(5)

    def sign_key_uid(
        self, signkey, keyid, uid, cert_level=0, digest='SHA256',
        minimize=False
    ):
        """Sign the given uid on the specified key.

        ``signkey``
            ID of the key to be used for signing.  The key must be
            present on both the secret keyring and this GnuPG's
            public keyring.
        ``keyid``
            ID of the key to be signed.
        ``uid``
            Integer uid.  Note that uids count from ZERO.

        """
        # TODO export to temporary keyring to sign
        # TODO configure cert-digest-algo
        self.uid = uid
        self.state = self.STATE_MINIMIZE if minimize else self.STATE_INIT
        p = self._popen([
            '--command-fd', '0',
            '--status-fd', '1',
            '--default-cert-level', str(cert_level),
            '--no-ask-cert-level',
            '--cert-digest-algo', 'SHA256',
            '--local-user', signkey,
            '--edit-key', keyid
        ])
        while p.poll() is None:
            line = p.stdout.readline().strip()
            logger.debug(line)
            response = self._get_response(line)
            if response is not None:
                logger.debug(response)
                p.stdin.write(response + '\n')
            # TODO set a max number of iteration and die if
            # exceeded, so we don't get stuck in loop?
        logger.info('gpg process exited code {}'.format(p.returncode))
        if p.returncode != 0:
            raise RuntimeError("Bad passphrase")

    def _get_response(self, line):
        msg = line[9:]  # strip "[GNUPG:] "
        lut = {
            "GOT_IT": None,
            "GET_LINE keyedit.prompt": self._get_line_keyedit_prompt,
            "GET_LINE sign_uid.class": '',  # accept default
            "GET_BOOL sign_uid.okay": 'y',
            # "USERID_HINT <full user id>"
            # "NEED_PASSPHRASE <long keyid> <longkeyid> 1 0"
            # "BAD_PASSPHRASE <long keyid>" <-- long keyid; exit code 2
            # "ALREADY_SIGNED <long keyid>" <-- exit code 0
        }
        try:
            response = lut[msg]
        except KeyError:
            logger.debug('ignore unrecognised message: {!r}'.format(msg))
            return None
        return response() if callable(response) else response

    def _get_line_keyedit_prompt(self):
        if self.state == self.STATE_MINIMIZE:
            self.state = self.STATE_INIT
            return 'minimize'
        if self.state == self.STATE_INIT:
            self.state = self.STATE_UID
            return '{}'.format(self.uid + 1)
        elif self.state == self.STATE_UID:
            self.state = self.STATE_SIGN
            return 'sign'
        elif self.state == self.STATE_SIGN:
            self.state = self.STATE_SAVE
            return 'save'
        else:
            raise RuntimeError('gpg state violation')



# This file is part of gcaff
# Copyright (C) 2013 Fraser Tweedale
#
# gcaff is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import shutil
import tempfile

import gtk

from . import gpg
from . import ui



def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--keyring', type=argparse.FileType(), required=True,
        help='keyring containing keys to be signed')
    parser.add_argument('--logging', default='WARNING',
        help='set log level')

    args = parser.parse_args()

    level = getattr(logging, args.logging.upper(), 'WARNING')
    logging.basicConfig(level=level)

    homegpg = gpg.GnuPG()

    tmpgpgdir = tempfile.mkdtemp()
    tmpgpg = gpg.GnuPG(tmpgpgdir)
    tmpgpg.import_keys(args.keyring.read(), minimal=True)
    # TODO import chosen signing key into tmpgpg

    window = ui.SigningAssistant(homegpg, tmpgpg)
    window.show_all()
    gtk.main()

    logging.warn('remove tmpgpgdir: {}'.format(tmpgpgdir))
    shutil.rmtree(tmpgpgdir)

if __name__ == '__main__':
    main()



# This file is part of gcaff
# Copyright (C) 2013 Fraser Tweedale
#
# gcaff is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Generate and send PGP/MIME messages.

Note that the "uid" in subroutines that take such an argument is a
full uid string.  Selecting a uid for uids that do not have an email
address is outside the scope of this module.

"""

import email.encoders
import email.mime.application
import email.mime.multipart
import email.mime.text
import logging
import smtplib

from . import version

logger = logging.getLogger(__name__)

TEMPLATE = u"""
Hi,

please find attached the user id

  {signee_uid}

of your key {signee_keyid} signed by me.

If you have multiple user ids, I sent the signature for each user id
separately to that user id's associated email address. You can
import the signatures by running each through `gpg --import`.

Note that I did not upload your key to any keyservers. If you want
this new signature to be available to others, please upload it
yourself.  With GnuPG this can be done using

  gpg --keyserver pool.sks-keyservers.net --send-key {signee_keyid}

If you have any questions, don't hesitate to ask.

Regards,

{signer_name}
"""


def create_message(
    gpg,
    payload,
    _from,
    _tos,
    signer_name,
    signee_keyid,
    signee_uid,
):
    """Create and return a PGP/MIME message and sender and recipient details.

    The return value is a 3-tuple consisting of:

    * "From" address
    * list of "To" addresses (which contains just the signee's address)
    * message as a bytestring

    An accumulation of such tuples is precisely the argument format
    required by ``send_messages``.

    """
    cleartext = _create_message(signee_keyid, signee_uid, signer_name, payload)
    msg = _encrypt_message(gpg, signee_keyid, cleartext)
    msg['From'] = _from
    msg['To'] = ', '.join(_tos)
    msg['Subject'] = 'Your signed PGP key 0x{}'.format(signee_keyid)
    msg['User-Agent'] = version.USER_AGENT
    return _from, _tos, msg.as_string()

def send_messages(msgs):
    """Generator that sends emails and yields progress."""
    server = smtplib.SMTP('localhost')
    for i, (_from, _tos, msg) in enumerate(msgs):
        server.sendmail(_from, _tos, msg)
        yield i + 1
    server.quit()

def _create_message(signee_keyid, signee_uid, signer_name, payload):
    text_plain = email.mime.text.MIMEText(
        TEMPLATE.format(
            signee_keyid=signee_keyid,
            signee_uid=signee_uid,
            signer_name=signer_name
        ),
        'plain',
        'utf-8'
    )
    application_pgp_keys = email.mime.application.MIMEApplication(
        payload,
        'pgp-keys',
        email.encoders.encode_7or8bit   
    )
    multipart = email.mime.multipart.MIMEMultipart('mixed')
    multipart.attach(text_plain)
    multipart.attach(application_pgp_keys)
    return str(multipart)

def _encrypt_message(gpg, keyid, cleartext):
    application_pgp_encrypted = email.mime.application.MIMEApplication(
        b'Version: 1',
        'pgp-encrypted',
        email.encoders.encode_7or8bit
    )
    application_octet_stream = email.mime.application.MIMEApplication(
        gpg.encrypt(keyid, cleartext),
        'octet-stream',
        email.encoders.encode_7or8bit
    )
    multipart = email.mime.multipart.MIMEMultipart(
        'encrypted',
        protocol='application/pgp-encrypted'
    )
    multipart.attach(application_pgp_encrypted)
    multipart.attach(application_octet_stream)
    return multipart



from setuptools import setup, find_packages

from gcaff import version

with open('README.rst') as f:
    readme = f.read()

setup(
    name='gcaff',
    version=version.VERSION,
    author='Fraser Tweedale',
    author_email='frase@frase.id.au',
    description='graphical OpenPGP signing assistant',
    long_description=readme,
    url='https://github.com/frasertweedale/gcaff',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 2.7',
        'Topic :: Communications :: Email',
        'Topic :: Security :: Cryptography',
    ],
    packages=find_packages(),
    entry_points={'console_scripts': ['gcaff = gcaff.__main__:main']},
)



# This file is part of gcaff
# Copyright (C) 2013 Fraser Tweedale
#
# gcaff is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import contextlib
import logging
import multiprocessing
import os
import tempfile

import gobject
import gtk

from . import gpg
from . import mail
from . import version

logger = logging.getLogger(__name__)


class SigningAssistant(gtk.Assistant):
    def __init__(self, usergpg, tmpgpg):
        super(SigningAssistant, self).__init__()
        self.set_title(version.USER_AGENT)
        self.connect('delete-event', gtk.main_quit)
        self.connect('cancel', gtk.main_quit)
        self.connect('close', gtk.main_quit)
        self.connect('prepare', self.on_prepare)

        self.signing_keys_by_keyid = {
            key.keyid: key
            for key in usergpg.list_secret_keys()
        }
        signing_keys = self.signing_keys_by_keyid.values()

        self.signing_keys = []
        self.signing_key_page = SigningKeyPage(signing_keys)
        self.append_page(self.signing_key_page)
        self.set_page_type(self.signing_key_page, gtk.ASSISTANT_PAGE_CONTENT)
        self.set_page_title(self.signing_key_page, 'Select the signing key(s).')
        self.signing_key_page.connect(
            'signing-keys-changed',
            self.on_signing_keys_changed
        )

        keys = (
            # FIXME remove constraint
            key for key in tmpgpg.list_keys()[:15]
            if key not in signing_keys
        )

        self.uid_selectors = []
        for key in keys:
            # TODO skip signing key
            uid_selector = UidSelector(key)
            self.append_page(uid_selector)
            self.set_page_type(uid_selector, gtk.ASSISTANT_PAGE_CONTENT)
            self.set_page_title(uid_selector, key.human_fingerprint())
            self.uid_selectors.append((key, uid_selector))

        confirm_page = ConfirmPage()
        self.append_page(confirm_page)
        self.set_page_type(confirm_page, gtk.ASSISTANT_PAGE_CONFIRM)

        # path to signature file
        self.sigfile = None

        self.progress_page = ProgressPage(usergpg, tmpgpg)
        self.append_page(self.progress_page)
        self.set_page_type(self.progress_page, gtk.ASSISTANT_PAGE_PROGRESS)
        self.progress_page.connect('sign-complete', self.on_sign_complete)
        self.progress_page.connect('send-complete', self.on_send_complete)

        summary_page = SummaryPage(self.progress_page)
        self.append_page(summary_page)
        self.set_page_type(summary_page, gtk.ASSISTANT_PAGE_SUMMARY)

    def _selected_uids(self):
        selected_uids = (
            (key, uid_selector.selected_uids())
            for key, uid_selector in self.uid_selectors
        )
        return [(key, sel) for key, sel in selected_uids if sel]

    def on_signing_keys_changed(self, page, keyids):
        self.signing_keys = \
            [self.signing_keys_by_keyid[keyid] for keyid in keyids]
        self.set_page_complete(self.signing_key_page, bool(keyids))

    def on_send_complete(self, page):
        self.set_page_complete(self.progress_page, True)

    def on_sign_complete(self, page, sigfile):
        self.sigfile = sigfile

    @staticmethod
    def on_prepare(assistant, page, *args):
        on_complete = lambda: assistant.set_page_complete(page, True)
        if isinstance(page, ProgressPage):
            page.sign_and_send(
                assistant.signing_keys,
                assistant._selected_uids()
            )
        elif not isinstance(page, SigningKeyPage):
            assistant.set_page_complete(page, True)


class SigningKeyPage(gtk.VBox):
    __gsignals__ = {
        'signing-keys-changed':
            (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (object,)),
    }

    def __init__(self, keys):
        super(SigningKeyPage, self).__init__()

        self.model = gtk.ListStore(bool, str, str, str)
        for key in keys:
            self.model.append((
                False,
                key.length,
                key.human_algorithm(),
                key.keyid,
            ))

        treeview = gtk.TreeView(self.model)
        self.pack_start(treeview)

        toggle_renderer = gtk.CellRendererToggle()
        toggle_renderer.connect('toggled', self.on_toggle, self.model)
        use_col = gtk.TreeViewColumn('Sign with?')
        use_col.pack_start(toggle_renderer, True)
        use_col.set_attributes(toggle_renderer, active=0)
        treeview.append_column(use_col)

        text_renderer = gtk.CellRendererText()
        length_col = gtk.TreeViewColumn('Length')
        length_col.pack_start(text_renderer, True)
        length_col.set_attributes(text_renderer, text=1)
        treeview.append_column(length_col)

        text_renderer = gtk.CellRendererText()
        algo_col = gtk.TreeViewColumn('Algorithm')
        algo_col.pack_start(text_renderer, True)
        algo_col.set_attributes(text_renderer, text=2)
        treeview.append_column(algo_col)

        text_renderer = gtk.CellRendererText()
        keyid_col = gtk.TreeViewColumn('Key ID')
        keyid_col.pack_start(text_renderer, True)
        keyid_col.set_attributes(text_renderer, text=3)
        treeview.append_column(keyid_col)

    def on_toggle(self, cell, path, model):
        model[path][0] = not model[path][0]
        signing_keys = [keyid for checked, _, _, keyid in model if checked]
        self.emit('signing-keys-changed', signing_keys)


class UidSelector(gtk.VBox):
    """Select UIDs on a key."""
    def __init__(self, key):
        super(UidSelector, self).__init__(spacing=10)
        self.set_border_width(10)

        self.cert_level = 0
        frame = gtk.Frame('Certification level')
        hbox = gtk.HBox()
        button_box = gtk.VButtonBox()
        button_box.set_border_width(5)
        button_box.set_spacing(-10)
        radio0 = gtk.RadioButton(None, '(0) I will not answer.')
        radio0.connect('clicked', self.on_radio_clicked, 0)
        radio1 = gtk.RadioButton(radio0, '(1) I have not checked at all.')
        radio1.connect('clicked', self.on_radio_clicked, 1)
        radio2 = gtk.RadioButton(radio0, '(2) I have done casual checking.')
        radio2.connect('clicked', self.on_radio_clicked, 2)
        radio3 = gtk.RadioButton(
            radio0,
            '(3) I have done very careful checking.'
        )
        radio3.connect('clicked', self.on_radio_clicked, 3)
        for radio in (radio0, radio1, radio2, radio3):
            button_box.add(radio)
        hbox.pack_start(button_box, expand=False)
        frame.add(hbox)
        self.pack_start(frame, expand=False)

        self.OFFSET = dict(zip(
            (
                'INDEX',
                'SELECTED',
                'SELECTABLE',
                'NOT_SELECTABLE',
                'TEXT',
                'PHOTO',
            ),
            range(6)
        ))
        self.model = gtk.ListStore(int, bool, bool, bool, str, gtk.gdk.Pixbuf)

        self.uids = key.uids
        for i, uid in enumerate(key.uids):
            if isinstance(uid, gpg.TextUid):
                label = uid.data
                pixbuf = None
            else:
                label = None
                with tempfile.NamedTemporaryFile() as f:
                    f.write(uid.data)
                    f.flush()
                    pixbuf = gtk.gdk.pixbuf_new_from_file(f.name)
            self.model.append((
                i,
                False,
                uid.signable(),
                not uid.signable(),
                label,
                pixbuf,
            ))

        treeview = gtk.TreeView(self.model)

        sign_col = gtk.TreeViewColumn('Sign?')
        toggle_renderer = gtk.CellRendererToggle()
        sign_col.pack_start(toggle_renderer, True)
        sign_col.set_attributes(
            toggle_renderer,
            activatable=self.OFFSET['SELECTABLE'],
            active=self.OFFSET['SELECTED']
        )
        toggle_renderer.connect('toggled', self.on_toggle, self.model)
        treeview.append_column(sign_col)

        uid_col = gtk.TreeViewColumn('UID')
        text_renderer = gtk.CellRendererText()
        uid_col.pack_start(text_renderer, True)
        uid_col.set_attributes(
            text_renderer,
            text=self.OFFSET['TEXT'],
            strikethrough=self.OFFSET['NOT_SELECTABLE']
        )
        pixbuf_renderer = gtk.CellRendererPixbuf()
        uid_col.pack_start(pixbuf_renderer, True)
        uid_col.set_attributes(pixbuf_renderer, pixbuf=self.OFFSET['PHOTO'])
        treeview.append_column(uid_col)
        # TODO column showing validity

        self.pack_start(treeview, expand=False)

    def on_toggle(self, cell, path, model):
        model[path][self.OFFSET['SELECTED']] \
            = not model[path][self.OFFSET['SELECTED']]

    def on_radio_clicked(self, button, cert_level):
        self.cert_level = cert_level

    def selected_uids(self):
        uids = []
        for row in self.model:
            if row[self.OFFSET['SELECTED']]:
                uids.append((
                    row[self.OFFSET['INDEX']],
                    self.uids[row[self.OFFSET['INDEX']]],
                    self.cert_level
                ))
        return uids


class ConfirmPage(gtk.HBox):
    def __init__(self):
        super(ConfirmPage, self).__init__()
        text = '''Sign and email all selected UIDs?'''
        self.pack_start(gtk.Label(text))


def sign_and_send(tmpgpg, uids, conn):
    """Sign and send signatures, sending results into the pipe."""
    n = len(uids) * 2
    c = 0
    signatures = []
    for i, (signargs, _) in enumerate(uids):
        logger.info('signing key, signargs={}'.format(signargs))
        try:
            tmpgpg.sign_key_uid(*signargs, minimize=True)
            signature = tmpgpg.export_keys([signargs[1]])
            logger.info('got signature: {}'.format(signature))
        except:
            logger.exception('error signing key: {}'.format(signargs))
            signature = None
        signatures.append(signature)
        c += 1
        conn.send((
            c / n,
            (True, None) if signature is not None else (False, 'sig fail')
        ))

    sigfile = _write_signatures(signatures)
    logger.warn('wrote signatures to file: {}'.format(sigfile))
    conn.send((None, sigfile))

    def _create_message(sig, mailargs):
        try:
            return mail.create_message(tmpgpg, sig, *mailargs)
        except Exception as e:
            e.mailargs = mailargs
            return e

    emails = [
        _create_message(sig, mailargs) if sig else None
        for (_, mailargs), sig in zip(uids, signatures)
    ]
    good_emails = filter(lambda x: not isinstance(x, Exception), emails)
    bad_emails = filter(lambda x: isinstance(x, Exception), emails)
    logger.info('failed to create {} emails'.format(len(bad_emails)))
    for e in bad_emails:
        c += 1
        conn.send((c / n, (False, e)))
    logger.info('sending {} emails'.format(len(good_emails)))
    for i in mail.send_messages(good_emails):
        c += 1
        conn.send((c / n, (True, None)))
    conn.send((1.0, (True, True)))

@contextlib.contextmanager
def _closing_mkstemp():
    fd, name = tempfile.mkstemp()
    try:
        yield fd, name
    finally:
        os.close(fd)

def _write_signatures(signatures):
    with _closing_mkstemp() as (fd, name):
        os.write(fd, '\n'.join(sig for sig in signatures if sig))
        return name


class ProgressPage(gtk.VBox):
    __gsignals__ = {
        'sign-complete':
            (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (str,)),
        'send-complete':
            (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ()),
    }

    def __init__(self, usergpg, tmpgpg):
        super(ProgressPage, self).__init__()
        self.usergpg = usergpg
        self.tmpgpg = tmpgpg
        self.progress_bar = gtk.ProgressBar()
        self.pack_start(self.progress_bar)

        self.notice = gtk.Label()
        self.notice.set_selectable(True)
        self.pack_start(self.notice)

        self.connect('sign-complete', self.on_sign_complete)

    def sign_and_send(self, signing_keys, uids):
        sign_ids = [key.keyid for key in signing_keys]
        logger.debug('import signing keys from user gpg: {}'.format(sign_ids))
        self.tmpgpg.import_keys(
            self.usergpg.export_keys(sign_ids, minimal=True)
        )

        logger.debug('normalising keylist: {}'.format(uids))
        spec = [
            (
                (signkey.keyid, key.keyid, uid_index, cert_level),
                (
                    signkey.email()[0] if signkey.email() else None,
                    key.uid_email(uid_index),
                    signkey.name()[0] if signkey.name() else None,
                    key.keyid,
                    str(key.uids[uid_index])
                ),
            )
            for signkey in signing_keys
            for key, keyuids in uids
            for uid_index, uid, cert_level in keyuids
        ]
        logger.debug('sign and send: {}'.format(spec))
        parent_conn, child_conn = multiprocessing.Pipe()
        p = multiprocessing.Process(
            target=sign_and_send,
            args=(self.tmpgpg, spec, child_conn)
        )
        p.start()
        gobject.timeout_add(200, self.on_check, parent_conn)

    def on_check(self, conn):
        while conn.poll():
            fraction, payload = conn.recv()
            if fraction is None:
                # status message
                logger.info('SIGNING COMPLETE')
                self.emit('sign-complete', payload)
            else:
                # progress message
                self.progress_bar.set_fraction(fraction)
                if not payload[0]:  # error
                    logger.error(repr(payload[1]))
                if fraction >= 1:
                    self.emit('send-complete')
                    return False
        return True

    def on_sign_complete(self, page, sigfile):
        logger.debug('GOT SIGFILE {}'.format(sigfile))
        self.notice.set_label(
            "Signatures have been written to {} ."
            .format(sigfile)
        )


class SummaryPage(gtk.VBox):
    def __init__(self, progress_page):
        super(SummaryPage, self).__init__()
        text = 'Signing and emailing completed.'
        self.pack_start(gtk.Label(text))

        self.import_help_label = gtk.Label()
        self.import_help_label.set_line_wrap(True)
        self.import_help_label.set_selectable(True)
        self.pack_start(self.import_help_label)

        progress_page.connect('sign-complete', self.on_sign_complete)

    def on_sign_complete(self, page, sigfile):
        msg = (
            'Signatures have been written to <tt>{0}</tt>.  You can import '
            'them into your keyring by issuing the following command at a '
            'terminal:\n'
            '\n'
            '<tt>gpg --import &lt; {0}</tt>'
        ).format(sigfile)
        self.import_help_label.set_markup(msg)
